/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import ws.ladder.exception.MultipleClassForPropertyException;
import ws.ladder.model.view.ContentViews;
import ws.ladder.service.AggregateManager;
import ws.ladder.service.DataAccessService;
import ws.ladder.util.ClassUtil;

public abstract class AbstractTypeNarrower implements TypeNarrower {

	private static Map<Class<?>, Map<String, Object>> subTypeProperties = new ConcurrentHashMap<Class<?>, Map<String, Object>>();

	private AggregateManager aggregateManager;

	public AggregateManager getAggregateManager() {
		return aggregateManager;
	}

	public void setAggregateManager(AggregateManager aggregateManager) {
		this.aggregateManager = aggregateManager;
	}

	@Override
	public List<String> getBasicAttributes(Class<?> clazz) {
		return getBasicAttributes(clazz, "");
	}

	private List<String> getBasicAttributes(Class<?> clazz, String baseName) {	
		// returns the BASIC information for this class
		clazz = aggregateManager.getTypeMapper().toReference(clazz);
		Type type = aggregateManager.getDAS().getType(clazz);

		List<String> attributeList = new ArrayList<String>();
		for(Property property: type.getProperties()) {
			if(property.isMany())
				continue;

			if(property.getType().isDataType()) {
				attributeList.add(baseName + property.getName());
				continue;
			}

			if( ((ExtendedType)property.getType()).isImmutable() )
				attributeList.addAll(getImmutableProperties(property.getType(), baseName + property.getName()));

			if( !property.isNullable() && !((ExtendedType)property.getType()).isImmutable() )
				attributeList.addAll( getInitializeProperties( baseName, property));
		}

		return attributeList;
	}	

	private List<String> getInitializeProperties(String baseName, Property p) {		
		ExtendedType requiredType = (ExtendedType) p.getType();
		baseName = baseName + p.getName();

		// Always add the id
		List<String> attributeList = new ArrayList<String>();
		if(p.isContainment()) {
			return getBasicAttributes(requiredType.getInstanceClass(), baseName + Settings.PATH_DELIMITER);
		} else {
			if(requiredType.getIdentifierProperty() != null)
				attributeList.add(baseName + Settings.PATH_DELIMITER + requiredType.getIdentifierProperty().getName());
			for(Property property: requiredType.getProperties()) {
				if( property == requiredType.getIdentifierProperty() )
					continue;

				if( ((ExtendedProperty)property).needsInitialization() && property.getType().isDataType())
					attributeList.add(baseName + Settings.PATH_DELIMITER + property.getName());
			}
		}

		return attributeList;
	}		

	private List<String> getImmutableProperties(Type category, String baseName) {
		List<String> attributeList = new ArrayList<String>();

		for(Property property: category.getProperties()) {
			if(!property.isMany() && property.getType().isDataType())
				attributeList.add(baseName + Settings.PATH_DELIMITER + property.getName());
		}

		return attributeList;
	}	

	/**
	 * This method is used to perform dynamic type narrowing. Subclasses override this method
	 * to provide custom behavior.
	 */
	@Override
	public Class<?> narrow(Object entity, String viewName) {
		Class<?> entityClass = ClassUtil.getUnEnhanced(entity.getClass());

		ContentViews.refresh(this);
		ContentViews.populateNarrowedClass(entityClass, this);		

		return ContentViews.getNarrowedClass(entityClass, viewName);
	}

	/**
	 * This method provides static type narrowing. Subclasses generally don't override
	 * this method.
	 */
	public Class<?> narrow(Class<?> entityClass, String propertyName) {
		Type entityType = null;

		TypeMapper typeMapper = aggregateManager.getTypeMapper();
		DataAccessService das = aggregateManager.getDAS();

		Class<?> referenceClass = entityClass;
		if(typeMapper.isDerived(entityClass))
			referenceClass = typeMapper.toReference(entityClass);

		entityType = das.getType(referenceClass);
		if(SimpleType.class.isAssignableFrom(entityType.getClass()))
			return entityType.getInstanceClass();

		return narrow(entityType, propertyName);
	}

	private Class<?> narrow(Type entityType, String propertyName) {
		if(entityType.getProperty(propertyName) != null)
			return entityType.getInstanceClass();
		else
			return findSubclass(entityType, propertyName);
	}

	private void populate(Type entityType) {
		if(!ExtendedType.class.isAssignableFrom(entityType.getClass()))
			return;

		ExtendedType type = (ExtendedType) entityType;
		TypeMapper typeMapper = aggregateManager.getTypeMapper();		

		// Find the property in the sub-class
		Map<String, Object> allEntityProperties = new HashMap<String, Object>();
		for(ExtendedType subType: type.getSubtypes()) {
			for(Property property: subType.getProperties()) {
				if(allEntityProperties.get(property.getName()) == null)
					allEntityProperties.put(property.getName(), subType.getInstanceClass());
				else { // multiple matches
					Object object = allEntityProperties.get(property.getName());
					if(Set.class.isAssignableFrom(object.getClass()))
						((Set) object).add(subType.getInstanceClass());
					else { // create a collection object and add to it the existing and the new class
						Set entityWithProperty = new HashSet();
						entityWithProperty.add(object);
						entityWithProperty.add(subType.getInstanceClass());
						allEntityProperties.put(property.getName(), entityWithProperty);
					}
				}
			}	
		}
		subTypeProperties.put(entityType.getInstanceClass(), allEntityProperties);
	}

	private Class<?> findSubclass(Type entityType, String propertyName) {
		if(subTypeProperties.get(entityType.getInstanceClass()) == null)
			populate(entityType);

		Map<String, Object> allEntityProperties = subTypeProperties.get(entityType.getInstanceClass());
		if(allEntityProperties.get(propertyName) == null)
			return null; // No sub-class has this property

		// Find the subclass
		Object object = allEntityProperties.get(propertyName);
		if(Class.class.isAssignableFrom(object.getClass()))
			return (Class<?>) object;

		// There is more than one match for the property throw an exception
		throw new MultipleClassForPropertyException(entityType, propertyName, object);
	}
}