/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.model;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.model.event.PropertyElement;
import ws.ladder.model.event.PropertyEvent;
import ws.ladder.model.operation.CloneOperation;
import ws.ladder.model.operation.ModifyOperation;
import ws.ladder.model.operation.QueryOperation;
import ws.ladder.model.operation.ReadOperation;
import ws.ladder.model.view.ViewProperty;
import ws.ladder.service.DataAccessService;
import ws.ladder.util.ClassUtil;
import ws.ladder.util.ObjectCreator;

public class ExtendedDataObjectImpl implements ExtendedDataObject, Node {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());
	private static final long serialVersionUID = 1L;

	// SDO Path related constants
	public static final String PATH_DELIMITER = "/";
	public static final String INDEX_FROM_0   = ".";
	public static final String INDEX_START    = "[";
	public static final String INDEX_END      = "]";	
	public static final String PATH_CONTAINER = "..";
	public static final String ATTR_DELIMITER = "=";	
	public static final String INSERT_OPERATOR = "+";  // insert at index for a set operation. Returns the element at index for a get operation.
	public static final String APPEND_OPERATOR = "<<"; // append at end of list for a set operation. Returns the last element in a GET operation.

	protected Object                             instance;                 // The object containing the actual data. This can be a JPA/Hibernate/Javabean object.
	protected Type                               type;                     // Represents the type of the entity, that holds the properties that make up this entity
	protected DataObject                         container;                // A reference to owning object of this entity in an aggregate
	protected Property                           containmentProperty;      // Represents the property referenced from the owning entity to this entity
	protected ChangeSummaryImpl                  changeSummary;            // Part of the SDO specification to track changes. Not needed unless need to make batch changes.
	protected ObjectCreator                      objectCreator;            // Cache to hold objects. Useful to set multiple references to an object and providing repeatable read capability.
	protected String                             objectPath;               // The path by which this object is referenced. This is PATH_DELIMITER for the root object.
	protected DependencyGraph                    dependencyGraph;          // Mechanism used to resolve dependencies and property order persistence related actions without causing data integrity conflicts.
	protected Map<Object, ExtendedDataObject>    aggregate;                // Map between source and data object of the aggregate
	protected boolean                            visited;                  // Used to prevent loops when traversing the graph

	// Attributes related to supporting the Node interface
	protected Edge           owner;                                       // collection owner, this is different from the container attribute, which refers to the collection object. The collection owner refers to the container of the container.
	protected Set<Edge>      outgoing = new HashSet<Edge>();              // The set of not-null attributes referenced BY this entity
	protected Set<Edge>      incoming = new HashSet<Edge>();              // The set of attributes referenced TO this entity
	protected String         aggregateName;                               // The name of the aggregate that this object is part of

	/* If supported by the persistence store, this helps the store to identify update conflicts. 
	 * This is different than the version attribute on the entity which is managed by the application/framework code
	 * 
	 * The persistence store takes care of converting the String to an internal representation as required by the store.
	 * For e.g., a byte[] to represent a vector clock in Riak
	 */
	protected String  versionSetByPersistenceStore;    
	protected boolean persistent;                                         // This flag is set to true, if the object was created from the persistence store
	protected boolean modified;                                           // This flag is set to true, if any fields or collections have been modified
	protected boolean readOnly;                                           // This flag is set to true, for read/query operations

	@Override
	public boolean isReadOnly() {
		return readOnly;
	}

	@Override
	public void setReadOnly(boolean readOnly) {
		this.readOnly = readOnly;
	}

	@Override
	public boolean isModified() {
		return modified;
	}

	@Override
	public void setModified(boolean modified) {
		this.modified = modified;
	}

	@Override
	public boolean isPersistent() {
		return persistent;
	}

	@Override
	public void setPersistent(boolean persistent) {
		this.persistent = persistent;
	}

	public Set<Edge> getOutgoing() {
		return outgoing;
	}

	public Set<Edge> getIncoming() {
		return incoming;
	}

	public void addIncomingEdge(Edge edge) {
		if(!incoming.contains(edge))
			incoming.add(edge);
	}

	public void removeIncomingEdge(Edge edge) {
		incoming.remove(edge);
	}

	public void addOutgoingEdge(Edge edge) {
		if(!outgoing.contains(edge))
			outgoing.add(edge);
	}

	public void removeOutgoingEdge(Edge edge) {
		outgoing.remove(edge);
	}

	/**
	 * Sets the aggregate name on all the aggregate parts
	 * @param aggregateName
	 * @return true if the seal was set successfully on all the parts
	 */
	public void setAggregateSeal(String aggregateName) {		
		if(visited)
			return;

		this.setAggregateName(aggregateName);
		visited = true;

		for(Edge edge: outgoing) {
			if(edge.isCascaded()) {
				if(edge.target == null)
					throw new RuntimeException("Dangling edge");
				edge.target.setAggregateSeal(aggregateName);
			}
		}
	}

	public void replaceInstance(Object newInstance) {
		// get the incoming edges and set the value to the new copy
		this.instance = newInstance;
		for(Edge edge: incoming) {
			((ExtendedProperty) edge.getProperty()).setValue(edge.source.getInstance(), edge.target.getInstance());
		}
	}

	/**
	 * 
	 * @param sourceOriginal can be any object within the source node
	 * @param targetOriginal should be the root node of the target scopeFinder
	 * @param propertyName
	 * @param isRequired
	 */
	public Edge addIncomingEdge(Node source, Property property) {
		Edge edge = new Edge(source, this, property);
		edge.link();

		if(edge.getProperty() != null && edge.getProperty().isMany())
			owner = edge;

		return edge;
	}

	public boolean isCollectionNode() {
		return owner != null;
	}

	public Edge getOwner() {
		return owner;
	}

	@Override
	public DependencyGraph getDependencyGraph() {
		return dependencyGraph;
	}

	public ExtendedDataObjectImpl(Type type, DataObject container, Property containmentProperty, ObjectCreator objectCreator) {
		this.type                = type;
		this.container           = container;
		this.containmentProperty = containmentProperty;
		this.changeSummary       = new ChangeSummaryImpl(this);
		this.objectCreator       = objectCreator;
		this.aggregate           = new IdentityHashMap<Object, ExtendedDataObject>();

		if(isRoot()) 
			dependencyGraph = new DependencyGraph(objectCreator);

		setReadOnly(container != null ? ((ExtendedDataObjectImpl) container).isReadOnly() : false);		

		setObjectPath();
	}

	@Override
	public boolean isRoot() {
		return this == getRootObject();
	}

	@Override
	public boolean isVisited() {
		return visited;
	}

	@Override
	public void setVisited(boolean visited) {
		this.visited = visited;
	}

	/**
	 * This method takes care of creating the containment graph and also any needed DataObject wrappers for the instance objects
	 */	
	@Override
	public void createAggregate() {
		aggregate = new IdentityHashMap<Object, ExtendedDataObject>();

		// Loop through the data object properties and if it is not a data type, then create a Data Object wrapper and recurse
		clearVisited();
		createWrapper(this);

		clearVisited();
	}

	@Override
	public boolean isAggregatePart(Object object) {
		if(isRoot()) {
			Object instance = ClassUtil.getInstance(object);
			return aggregate.containsKey(instance);
		} else
			return ((ExtendedDataObject)getRootObject()).isAggregatePart(object);
	}

	protected void clearVisited() {
		for(ExtendedDataObject dataObject: objectCreator.getDataObjects())
			dataObject.setVisited(false);
	}

	@Override
	public void addEntity(ExtendedDataObject entity) {
		Object id = entity.getIdentifierValue();
		if(id != null)
			getObjectCreator().addByEntityKey(getEntityKey(id, entity.getType()), entity);
	}

	@Override
	public void removeEntity(ExtendedDataObject entity) {
		Object id = entity.getIdentifierValue();
		if(id != null) 
			getObjectCreator().removeByEntityKey(getEntityKey(id, entity.getType()));
	}

	/**
	 * Get the derived data object based off a reference data object
	 * <tt>EntityKey</tt>
	 */
	@Override
	public ExtendedDataObject getEntity(ExtendedDataObject entity) {
		Object id = entity.getIdentifierValue();
		if(id != null)
			return getByEntityKey(id, entity.getType());

		return null;
	}

	private EntityKey getEntityKey(Object id, Type type) {
		if(id == null)
			return null;

		if(!ExtendedType.class.isAssignableFrom(type.getClass()))
			throw new IllegalArgumentException("type has to refer to a Data Object");

		Type rootEntityType = ((ExtendedType)type).getRootEntityType();
		EntityKey key = getObjectCreator().getTypeMapper().getEntityKey(id.toString(), rootEntityType); 

		return key;
	}

	public ExtendedDataObject getByEntityKey(Object id, Type type) {
		return getObjectCreator().getByEntityKey(getEntityKey(id, type));		
	}

	protected void createWrapper(ExtendedDataObject parent) {

		for(ExtendedDataObject child: parent.getList()) {
			((Node)child).addIncomingEdge((Node) parent, null);
			createWrapper(child);
		}

		for(Property property: parent.getType().getProperties()) {	
			if(!((ExtendedProperty) property).isDataType()) {
				Object propertyInstance = ((ExtendedProperty)property).getValue(parent);
				if(propertyInstance == null)
					continue;

				Object target = objectCreator.getExistingDataObject(propertyInstance);
				if(target != null && !ExtendedDataObject.class.isAssignableFrom(target.getClass()))
					throw new IllegalStateException("Property refers to a DataObject, but the object is not a DataObject");

				ExtendedDataObject child = null;
				if(target != null)
					child = (ExtendedDataObject) target;
				else {
					ExtendedDataObject container = parent;
					Property containmentProperty = property;
					if(!property.isContainment()) {
						container = null;
						containmentProperty = null;
					}
					child = objectCreator.createDataObject(propertyInstance, property.getType(), container, containmentProperty);
				}
				((Node)child).addIncomingEdge((Node) parent, property);

				if(child.isVisited())
					continue;
				else
					child.setVisited(true);

				if(property.isContainment()) 
					aggregate.put(propertyInstance, child);
				else
					continue;

				createWrapper(child);				
			}
		}
	}

	@Override
	public ExtendedProperty getCollectionKeyProperty() {
		Type type = ((ExtendedProperty)containmentProperty).getElementType();
		if(ExtendedType.class.isAssignableFrom(type.getClass()))
			return (ExtendedProperty) ((ExtendedType)type).getCollectionUserKey();

		return null;
	}

	protected void setObjectPath() {
		StringBuffer path = (container == null) ? new StringBuffer(PATH_DELIMITER) : new StringBuffer(((ExtendedDataObject)container).getObjectPath());

		// Check the association type

		// Handle collections
		if(containmentProperty != null && containmentProperty.isMany() && !((ExtendedProperty)containmentProperty).getElementType().isDataType()) {
			// If this is an indexable collection we should be using the index. Unfortunately this is not possible, since 
			// a list can have multiple indexes point to the same item.

			Property collectionUserKey = getCollectionKeyProperty();			
			path.append(PATH_DELIMITER).append(containmentProperty.getName()).append(INDEX_START);
			if( collectionUserKey != null ) {
				path.append(collectionUserKey.getName()).append(ATTR_DELIMITER);
				path.append(this.get(collectionUserKey));
			} else { // fallback to id
				Type elementType = ((ExtendedProperty)containmentProperty).getElementType();
				Property identifier = ((ExtendedType)elementType).getIdentifierProperty();
				path.append(identifier.getName()).append(ATTR_DELIMITER); 
				path.append(this.get(identifier));
			}
			objectPath = path.append(INDEX_END).toString();
		} else if(containmentProperty != null) {
			if(!path.toString().endsWith(PATH_DELIMITER))
				path.append(PATH_DELIMITER);
			objectPath = path.append(containmentProperty.getName()).toString(); 
		} else 
			objectPath = path.toString();
	}

	@Override
	public String getObjectPath() {
		return objectPath;
	}

	@Override
	public Object getInstance() {
		return instance;
	}

	@Override
	public void setInstance(Object instance) {
		if(instance != null && this.getContainmentProperty() != null) {
			ExtendedProperty property = (ExtendedProperty) this.getContainmentProperty();
			if( (property.isSet() && !Set.class.isAssignableFrom(instance.getClass())) ||
					(property.isList() && !List.class.isAssignableFrom(instance.getClass())) ||
					(property.isMap() && !Map.class.isAssignableFrom(instance.getClass())) ||
					( !property.isSet() && !property.isList() && !property.isMap() && !type.getInstanceClass().isAssignableFrom(instance.getClass()))
					)
				throw new IllegalArgumentException("The DataObject instance class is not compatible with its type");
		}

		this.instance = instance;
	}

	private boolean isIndexOperation(String indexStr) {
		if(indexStr.trim().endsWith(INSERT_OPERATOR) || indexStr.trim().equals(APPEND_OPERATOR))
			return true;

		return false;
	}

	private Object getIndexObject(String indexStr, String propertyName) {
		if(indexStr.trim().endsWith(INSERT_OPERATOR)) {
			int index = Integer.valueOf(indexStr.substring(0, indexStr.indexOf(INSERT_OPERATOR)).trim() );
			return ((List<?>)get(getType().getProperty(propertyName))).get(index);
		} else if(indexStr.trim().equals(APPEND_OPERATOR)) {
			List<?> listObj = (List<?>)get(getType().getProperty(propertyName));
			return listObj.get(listObj.size()-1);
		}

		return null;
	}

	@SuppressWarnings("unchecked")
	private void setIndexObject(String indexStr, String propertyName, Object value) {
		List<Object> valueList = (List<Object>) get(propertyName);		
		if(indexStr.trim().endsWith(INSERT_OPERATOR)) {
			int index = Integer.valueOf(indexStr.substring(0, indexStr.indexOf(INSERT_OPERATOR)).trim() );
			valueList.add(index, value);			

		} else if(indexStr.trim().equals(APPEND_OPERATOR))
			valueList.add(value);
	}	

	/**
	 *  SDO Path expression. Enhancements made for LWSDO.
	 * 
	 *   path ::= (scheme ':')? '/'? (step '/')*   step
	 *   scheme ::= [^:]+
	 *   step ::= '@'? property
	 *          | property '[' index_from_1 ']'
	 *          | property '.' index_from_0
	 *          | reference '[' attribute '=' value ']'
	 *          | ".."
	 *   property ::= NCName ;; may be simple or complex type
	 *   attribute ::= NCName ;; must be simple type
	 *   reference :: NCName ;; must be DataObject type
	 *   index_from_0 ::= Digits ('+')? | "<<"
	 *   index_from_1 ::= NotZero (Digits)? ('+')? | "<<"
	 *   value ::= Literal
	 *          | Number
	 *          | Boolean
	 *   Literal ::= '"' [^"]* '"'
	 *          | "'" [^']*   "'"
	 *   Number ::= Digits ('.' Digits?)?
	 *          | '.' Digits
	 *   Boolean ::= true
	 *          | false
	 *   NotZero ::= [1-9]
	 *   Digits ::= [0-9]+
	 *   
	 *   ;; leading '/' begins at the root
	 *   ;; ".." is the containing DataObject, using containment properties
	 *   ;; Only the last step have an attribute as the property
	 *   
	 *   The scheme is an extension mechanism for supporting additional path expressions 
	 *   in the future. No schema and a scheme of "sdo:" are equivalent, representing this syntax.
	 *   
	 *   If more than one property shares the same name, only the first is matched by the path expression, 
	 *   using property.name for name matching. If there are alias names assigned,
	 *   those are also used to match. Also, names including any of the special characters of the
	 *   syntax (./[]=@) are not accessible. Each step of the path before the last must return a
	 *   single DataObject. When the property is a Sequence, the values returned are those of the
	 *   getValue() accessor.
	 *   
	 * 
	 * @param path
	 * @return
	 */
	private Object getPathObject(String path) {

		if(path == null || "".equals(path))
			throw new IllegalArgumentException("Path cannot be empty or null");

		if(path.equals(PATH_CONTAINER))
			return getContainer();

		if(path.equals(PATH_DELIMITER)) { // root
			DataObject result = this;
			while(result.getContainer() != null)
				result = result.getContainer();
			return result;
		}

		// multiple components
		if(path.contains(PATH_DELIMITER)) {
			if(path.startsWith(PATH_DELIMITER)) {
				return ((DataObject)getPathObject("/")).get(path.substring(path.indexOf(PATH_DELIMITER)+1));
			} else {
				String firstComponent = path.substring(0, path.indexOf(PATH_DELIMITER));
				String remainingPath = path.substring(path.indexOf(PATH_DELIMITER)+1);
				// Get the property name related to the first path component
				return ((DataObject)getPathObject(firstComponent)).get(remainingPath);
			}
		}

		// single component
		// List element indexed from 0
		if(path.contains(INDEX_FROM_0)) {
			String propertyName = path.substring(0, path.indexOf(INDEX_FROM_0));
			String indexStr = path.substring(path.indexOf(INDEX_FROM_0)+1);
			try {
				int index = Integer.valueOf(indexStr);
				return ((List<?>)get(getType().getProperty(propertyName))).get(index);
			} catch (NumberFormatException e) {
				if(isIndexOperation(indexStr))
					return getIndexObject(indexStr, propertyName);

				// Map collection
				Property property = getType().getProperty(propertyName);
				if(Map.class.isAssignableFrom(property.getType().getInstanceClass())) {
					Map<?, ?> values = (Map<?, ?>) get(property);
					return values.get(indexStr);					
				}
			}
		}

		// List element indexed from 1
		if(path.contains(INDEX_START)) {
			String propertyName = path.substring(0, path.indexOf(INDEX_START));
			String indexStr = path.substring(path.indexOf(INDEX_START)+1, path.indexOf(INDEX_END));

			try {
				int index = Integer.valueOf(indexStr);
				return ((List<?>)getPathObject(propertyName)).get(index-1);
			} catch (NumberFormatException e) {
				if(isIndexOperation(indexStr))
					return getIndexObject(indexStr, propertyName);

				// List element indexed by attribute value
				return getByPropertyFilter(path);
			}
		}		

		return get(getType().getProperty(path));
	}

	@SuppressWarnings("unchecked")
	protected Object getByPropertyFilter(String path) {

		String propertyName = path.substring(0, path.indexOf(INDEX_START));
		String indexStr = path.substring(path.indexOf(INDEX_START)+1, path.indexOf(INDEX_END));

		if( ((ExtendedProperty)containmentProperty).isMap() && indexStr.indexOf(ATTR_DELIMITER) == -1) {
			Map<?, ?> values = (Map<?, ?>) instance;
			Object valueInstance = values.get(indexStr);
			return objectCreator.createDataObject(valueInstance, ((ExtendedProperty)containmentProperty).getElementType(), (ExtendedDataObject) container, containmentProperty);
		}

		// TODO: Check path validity 

		if(((ExtendedProperty)containmentProperty).getElementType().isDataType())
			throw new IllegalArgumentException("The property '" + propertyName + "' refers to a non-DataObject collection");

		if(!Collection.class.isAssignableFrom(containmentProperty.getType().getInstanceClass()))	
			throw new IllegalStateException("Expecting a collection instance");

		Collection<DataObject> values = (Collection<DataObject>) getList(containmentProperty);

		String attributeName = indexStr.substring(0, indexStr.indexOf(ATTR_DELIMITER));
		Property attributeProperty = ((ExtendedProperty)containmentProperty).getElementType().getProperty(attributeName);
		if(!attributeProperty.getType().isDataType())
			throw new IllegalArgumentException("Attribute '" + attributeName + "' cannot have a value of type DataObject");

		String attributeValue = indexStr.substring(indexStr.indexOf(ATTR_DELIMITER)+1);
		attributeValue = attributeValue.replaceAll("^\"|\"$|^'|'$", "");

		// TODO: ensure the string representation is according to the SDO spec

		// Retrieve by identifier if possible
		if( ((ExtendedType)((ExtendedProperty)containmentProperty).getElementType()).getIdentifierProperty() == attributeProperty ) {
			return getByEntityKey(attributeValue, ((ExtendedProperty)containmentProperty).getElementType());
		} else {
			for(DataObject dataObject: values) {
				if(dataObject.get(attributeProperty).toString().equals(attributeValue))
					return dataObject;
			}		
		}

		return null;
	}

	@Override
	public Object get(String path) {
		return getPathObject(path);
	}

	protected DataObject getDeepestContainer(String path) {
		if(path == null || "".equals(path))
			throw new IllegalArgumentException("Path cannot be empty or null");

		if(path.equals(PATH_CONTAINER) || path.equals(PATH_DELIMITER))
			throw new IllegalArgumentException("Path should refer to a property");

		// Get the container
		ExtendedDataObjectImpl container = this;
		if(path.contains(PATH_DELIMITER)) { // find the correct container
			String containerPath = path.substring(0, path.lastIndexOf(PATH_DELIMITER));
			container = (ExtendedDataObjectImpl) getPathObject(containerPath);
			path = path.substring(path.lastIndexOf(PATH_DELIMITER)+1);
		}	

		return container;
	}

	@Override
	public ExtendedDataObject createDataObject(Object id, Type instanceType, Property property) throws Exception {
		Object propertyInstance = objectCreator.createInstance(instanceType);
		if(!instanceType.isDataType() && ((ExtendedType)instanceType).getIdentifierProperty() != null )
			((ExtendedProperty)((ExtendedType)instanceType).getIdentifierProperty()).setValue(propertyInstance, id);

		ExtendedDataObject result = objectCreator.createDataObject(propertyInstance, instanceType, this, property);
		if(property != null)
			((ExtendedProperty)property).setValue(this, propertyInstance);

		return result;
	}

	@Override
	public ExtendedDataObject createDataObject(Object id, Type instanceType) throws Exception {
		Object propertyInstance = objectCreator.createInstance(instanceType);
		if(!instanceType.isDataType() && ((ExtendedType)instanceType).getIdentifierProperty() != null )
			((ExtendedProperty)((ExtendedType)instanceType).getIdentifierProperty()).setValue(propertyInstance, id);

		return objectCreator.createDataObject(propertyInstance, instanceType, null, null);
	}	

	@Override
	public void set(String propertyPath, Map<String, Object> propertyResult) throws Exception {

		// If we are setting a null value then nothing needs to be done
		if( propertyResult.get(ViewProperty.qualifyProperty(propertyPath) ) == null)
			return;

		// Since this builds the path for objects already persisted, the identifier value will not be null
		String[] pathSteps = propertyPath.split(Settings.PATH_DELIMITER_REGEX);
		ExtendedDataObject current = this;
		StringBuffer currentPath = new StringBuffer(ViewProperty.ROOT_PROPERTY_NAME);
		for(String step: pathSteps) {
			if(currentPath.length() > 0)
				currentPath.append(Settings.PATH_DELIMITER);
			currentPath.append(step);

			Property property = current.getInstanceProperty(step);
			if(property == null)
				throw new RuntimeException("Unable to resolve property: " + propertyPath);

			Object propertyDO = current.get(property);
			if(!property.isMany()) { 
				if(property.getType().isDataType() ) {
					// Populate the field and return the data object
					((ExtendedProperty)property).setValue(current, propertyResult.get(ViewProperty.qualifyProperty(propertyPath) ));
					return;
				}

				if( ((ExtendedType)property.getType()).isEmbedded() )
					continue; // Embedded objects are automatically created as part of its lifecycle owner

				if(propertyDO == null) {
					// Check if there is a data object with the given key
					// Get the identifier value
					Object idValue = null;
					if(((ExtendedType)property.getType()).getIdentifierProperty() != null) {
						idValue = propertyResult.get(currentPath + Settings.PATH_DELIMITER + ((ExtendedType)property.getType()).getIdentifierProperty().getName());
						propertyDO = getByEntityKey(idValue, property.getType());
					}
					if(propertyDO == null) { // create and set the instance object
						// check if we are narrowing
						String narrowToType = (String) propertyResult.get(currentPath + Settings.PATH_DELIMITER + ViewProperty.ENTITYNAME_ATTRIBUTE);
						ExtendedType objectType = (ExtendedType) property.getType();
						if(narrowToType != null)
							objectType = (ExtendedType) getObjectCreator().getDAS().getType(narrowToType);
						propertyDO = current.createDataObject(idValue, objectType, property);
					}
				}
				if(property.isContainment()) {
					((ExtendedDataObject)propertyDO).setContainer(current);
					((ExtendedDataObject)propertyDO).setContainmentProperty(property);
				}

				// Set the instance value in the container
				((ExtendedProperty)property).setValue(current, ((ExtendedDataObject)propertyDO).getInstance());

				current = (ExtendedDataObject) propertyDO;		
			} else  {

				if(propertyDO == null) { // create and set the collection/map object
					propertyDO = current.createDataObject(null, property.getType(), property);
				} else if(!ExtendedDataObject.class.isAssignableFrom(propertyDO.getClass()))
					propertyDO = objectCreator.createDataObject(propertyDO, property.getType(), current, property);
				current = (ExtendedDataObject) propertyDO;

				// Get the identifier value from the collection element
				Type elementType = ((ExtendedProperty)property).getElementType();
				Object idValue = null;
				if(((ExtendedType) elementType).getIdentifierProperty() != null) {
					idValue = propertyResult.get(currentPath + Settings.PATH_DELIMITER + ((ExtendedType) elementType).getIdentifierProperty().getName());
					propertyDO = getByEntityKey(idValue, elementType);

					// check flag to see if the containment should be set
					if(propertyDO != null && property.isContainment())
						((ExtendedDataObject)propertyDO).setContainer(current); // Containment property is null for a collection element					
				}

				if(propertyDO == null) { // create and set the instance object

					// Get the property instance if possible
					Object elementInstance = null;

					if( ((ExtendedProperty)property).isMap() ) {
						Object keyValue = propertyResult.get(currentPath + Settings.PATH_DELIMITER + ViewProperty.MAP_KEY_ATTRIBUTE);
						Map map = (Map) current.getInstance();
						elementInstance = map.get(keyValue);
					} else if ( ((ExtendedProperty)property).isList() || ((ExtendedProperty)property).isSet() ) {

						// If the collection element is an embeddable look for a user key otherwise throw a operation not supported. 
						// TODO: This check should be done in view validation	

						// TODO: performance optimization. Create an EntityKey object where Id is the collection owner id + property name + user key 
						Object colUserKeyValue = propertyResult.get(currentPath + Settings.PATH_DELIMITER + ViewProperty.COL_USERKEY_ATTRIBUTE);
						if(colUserKeyValue != null) {
							Collection collection = (Collection) current.getInstance();
							ExtendedProperty colUserKey = (ExtendedProperty) (ExtendedType.class.isAssignableFrom(elementType.getClass()) ? ((ExtendedType)elementType).getCollectionUserKey() : null);	
							for(Object obj: collection) {
								if( colUserKey != null && colUserKeyValue.equals(colUserKey.getValue(obj)) ) {
									elementInstance = obj;
									break;
								} 
							}
						}
					}

					if(elementInstance == null) {
						if(idValue != null)
							propertyDO = current.createDataObject(idValue, (ExtendedType) elementType, null);							
						else
							return; // Does not have a collection element
					} else
						// create the data object using the instance
						propertyDO = objectCreator.createDataObject(elementInstance, elementType, current, property);

					// check flag to see if the containment should be set
					if(property.isContainment())
						((ExtendedDataObject)propertyDO).setContainer(current); // Containment property is null for a collection element
				}

				// Add the element
				Object elementInstance = ((ExtendedDataObject)propertyDO).getInstance();				
				if( ((ExtendedProperty)property).isMap() ) {
					// If this is a map, get the key
					Object keyValue = propertyResult.get(currentPath + Settings.PATH_DELIMITER + ViewProperty.MAP_KEY_ATTRIBUTE);
					Map map = (Map) current.getInstance();
					map.put(keyValue, elementInstance);
				} else if ( ((ExtendedProperty)property).isList() ) {
					Object indexValue = propertyResult.get(currentPath + Settings.PATH_DELIMITER + ViewProperty.LIST_INDEX_ATTRIBUTE);					
					List list = (List) current.getInstance();
					int index = Integer.parseInt(indexValue.toString());
					if(index >= list.size() || list.get(index) != elementInstance)
						list.add(elementInstance);
				} else if ( ((ExtendedProperty)property).isSet() ) {
					Set set = (Set) current.getInstance();
					set.add(elementInstance);
				}				

				current = (ExtendedDataObject) propertyDO;	
			}
		}

		return;
	}

	@Override
	public Property getPropertyByPath(String path) {

		// Get the container
		ExtendedDataObjectImpl container = (ExtendedDataObjectImpl) getDeepestContainer(path);
		String propertyPath = path.substring(path.lastIndexOf(PATH_DELIMITER)+1);

		// case 1: List element indexed from 0
		if(propertyPath.contains(INDEX_FROM_0)) {
			String propertyName = propertyPath.substring(0, propertyPath.indexOf(INDEX_FROM_0));
			return container.getType().getProperty(propertyName.trim());
		}

		// case 2: List element indexed from 1
		if(propertyPath.contains(INDEX_START)) {
			String propertyName = propertyPath.substring(0, propertyPath.indexOf(INDEX_START));
			return container.getType().getProperty(propertyName.trim());
		}

		return container.getType().getProperty(propertyPath);
	}

	@Override
	public void set(String path, Object value) {
		// TODO: keep track of the modifications

		if(path == null || "".equals(path))
			throw new IllegalArgumentException("Path cannot be empty or null");

		if(path.equals(PATH_CONTAINER) || path.equals(PATH_DELIMITER))
			throw new IllegalArgumentException("Path should refer to a property");

		// Get the container
		ExtendedDataObjectImpl container = this;
		if(path.contains(PATH_DELIMITER)) { // find the correct container
			String containerPath = path.substring(0, path.lastIndexOf(PATH_DELIMITER));
			container = (ExtendedDataObjectImpl) getPathObject(containerPath);
			path = path.substring(path.lastIndexOf(PATH_DELIMITER)+1);
			container.set(path, value);
			return;
		}

		// Set the value for the property
		// Path now contains the last component
		if(path.contains(INDEX_FROM_0) || path.contains(INDEX_START)) { // List/attribute property setter
			// List element indexed from 0
			if(path.contains(INDEX_FROM_0)) {
				String propertyName = path.substring(0, path.indexOf(INDEX_FROM_0));
				String indexStr = path.substring(path.indexOf(INDEX_FROM_0)+1);
				try {
					int index = Integer.valueOf(indexStr);
					List<Object> valueList = (List<Object>) get(propertyName);
					valueList.set(index, value);
				} catch (NumberFormatException e) {
					Property property = getType().getProperty(propertyName);

					if(isIndexOperation(indexStr)) {
						setIndexObject(indexStr, propertyName, value);
					} else if(Map.class.isAssignableFrom(property.getType().getInstanceClass())) {
						Map<String, Object> values = (Map<String, Object>) get(propertyName);
						values.put(indexStr, value);					
					}
				}
				return;
			}		

			// List element indexed from 1
			if(path.contains(INDEX_START)) {
				String propertyName = path.substring(0, path.indexOf(INDEX_START));
				String indexStr = path.substring(path.indexOf(INDEX_START)+1, path.indexOf(INDEX_END));

				try {
					int index = Integer.valueOf(indexStr);
					List<Object> valueList = (List<Object>) get(propertyName);
					valueList.set(index-1, value);
				} catch (NumberFormatException e) {
					if(isIndexOperation(indexStr))
						setIndexObject(indexStr, propertyName, value);
					else
						// List element indexed by attribute value
						setByPropertyFilter(path, value);
				}
			}
		} else { // simple property setter
			ExtendedProperty property = (ExtendedProperty) getType().getProperty(path);
			property.setValue(getInstance(), value);
		}
	}

	protected void setByPropertyFilter(String path, Object value) {

		if(!DataObject.class.isAssignableFrom(value.getClass()))
			throw new IllegalArgumentException("value should be of type DataObject");

		String propertyName = path.substring(0, path.indexOf(INDEX_START));
		String indexStr = path.substring(path.indexOf(INDEX_START)+1, path.indexOf(INDEX_END));

		Property property = getType().getProperty(propertyName);
		if(Map.class.isAssignableFrom(property.getType().getInstanceClass()) && indexStr.indexOf(ATTR_DELIMITER) == -1) {
			Map<String, Object> values = (Map<String, Object>) get(property);
			values.put(indexStr, value);
			return;
		}		
		if(!DataObject.class.isAssignableFrom(property.getType().getInstanceClass()))
			throw new IllegalArgumentException("The property '" + propertyName + "' refers to a non-DataObject collection");
		List<DataObject> values = (List<DataObject>) get(property);

		String attributeName = indexStr.substring(0, indexStr.indexOf(ATTR_DELIMITER));
		Property attributeProperty = getType().getProperty(attributeName);
		if(!attributeProperty.getType().isDataType())
			throw new IllegalArgumentException("Attribute '" + attributeName + "' cannot have a value of type DataObject");

		String attributeValue = indexStr.substring(indexStr.indexOf(ATTR_DELIMITER)+1);
		attributeValue = attributeValue.replaceAll("^\"|\"$|^'|'$", "");

		// TODO: ensure the string representation is according to the SDO spec
		for(int i = 0; i < values.size(); i++) {
			DataObject dataObject = values.get(i);
			if(dataObject.get(attributeProperty).toString().equals(attributeValue)) {
				values.set(i, (DataObject) value);
				break;
			}
		}		
	}	

	@Override
	public boolean isSet(String path) {
		DataObject container = getDeepestContainer(path);
		Property property = getPropertyByPath(path);

		ChangeSummary.Setting setting = changeSummary.getOldValue(container, property);
		return setting.isSet();
	}

	@Override
	public void unset(String path) {
		DataObject container = getDeepestContainer(path);
		Property property = getPropertyByPath(path);
	}

	@Override
	public boolean getBoolean(String path) {
		Boolean result = (Boolean) getPathObject(path);

		return result;
	}

	@Override
	public byte getByte(String path) {
		Byte result = (Byte) getPathObject(path);

		return result;
	}

	@Override
	public char getChar(String path) {
		Character result = (Character) getPathObject(path);

		return result;
	}

	@Override
	public double getDouble(String path) {
		Double result = (Double) getPathObject(path);

		return result;
	}

	@Override
	public float getFloat(String path) {
		Float result = (Float) getPathObject(path);

		return result;
	}

	@Override
	public int getInt(String path) {
		Integer result = (Integer) getPathObject(path);

		return result;
	}

	@Override
	public long getLong(String path) {
		Long result = (Long) getPathObject(path);

		return result;
	}

	@Override
	public short getShort(String path) {
		Short result = (Short) getPathObject(path);

		return result;
	}

	@Override
	public byte[] getBytes(String path) {
		byte[] result = (byte[]) getPathObject(path);

		return result;
	}

	@Override
	public BigDecimal getBigDecimal(String path) {
		BigDecimal result = (BigDecimal) getPathObject(path);

		return result;
	}

	@Override
	public BigInteger getBigInteger(String path) {
		BigInteger result = (BigInteger) getPathObject(path);

		return result;
	}

	@Override
	public DataObject getDataObject(String path) {
		Object instanceObject = getPathObject(path);
		DataObject result = (DataObject) ((ExtendedDataObject)getRootObject()).getObjectCreator().getExistingDataObject(instanceObject);

		return result;
	}

	@Override
	public Date getDate(String path) {
		Date result = (Date) getPathObject(path);

		return result;
	}

	@Override
	public String getString(String path) {
		String result = getPathObject(path).toString();

		return result;
	}

	@Override
	public List<?> getList(String path) {
		return (new DataObjectList(this, getPropertyByPath(path))).list(null);
	}

	@Override
	public Sequence getSequence(String path) {
		Sequence result = (Sequence) getPathObject(path);

		return result;
	}

	@Override
	public void setBoolean(String path, boolean value) {
		set(path, value);
	}

	@Override
	public void setByte(String path, byte value) {
		set(path, value);
	}

	@Override
	public void setChar(String path, char value) {
		set(path, value);
	}

	@Override
	public void setDouble(String path, double value) {
		set(path, value);
	}

	@Override
	public void setFloat(String path, float value) {
		set(path, value);
	}

	@Override
	public void setInt(String path, int value) {
		set(path, value);
	}

	@Override
	public void setLong(String path, long value) {
		set(path, value);
	}

	@Override
	public void setShort(String path, short value) {
		set(path, value);
	}

	@Override
	public void setBytes(String path, byte[] value) {
		set(path, value);
	}

	@Override
	public void setBigDecimal(String path, BigDecimal value) {
		set(path, value);
	}

	@Override
	public void setBigInteger(String path, BigInteger value) {
		set(path, value);
	}

	@Override
	public void setDataObject(String path, DataObject value) {
		set(path, value);
	}

	@Override
	public void setDate(String path, Date value) {
		set(path, value);
	}

	@Override
	public void setString(String path, String value) {
		set(path, value);
	}

	@Override
	public void setList(String path, List<?> value) {
		set(path, value);
	}

	@Override
	public Object get(int propertyIndex) {
		Property property = (Property) getType().getProperties().get(propertyIndex);
		return get(property);
	}

	@Override
	public void set(int propertyIndex, Object value) {
		// TODO: track changes
		ExtendedProperty property = (ExtendedProperty) getType().getProperties().get(propertyIndex);
		property.setValue(this, value);
	}

	@Override
	public boolean isSet(int propertyIndex) {
		// TODO Auto-generated method stub
		return false;
	}

	@Override
	public void unset(int propertyIndex) {
		// TODO Auto-generated method stub

	}

	@Override
	public boolean getBoolean(int propertyIndex) {
		return (Boolean) get(propertyIndex);
	}

	@Override
	public byte getByte(int propertyIndex) {
		return (Byte) get(propertyIndex);
	}

	@Override
	public char getChar(int propertyIndex) {
		return (Character) get(propertyIndex);
	}

	@Override
	public double getDouble(int propertyIndex) {
		return (Double) get(propertyIndex);
	}

	@Override
	public float getFloat(int propertyIndex) {
		return (Float) get(propertyIndex);
	}

	@Override
	public int getInt(int propertyIndex) {
		return (Integer) get(propertyIndex);
	}

	@Override
	public long getLong(int propertyIndex) {
		return (Long) get(propertyIndex);
	}

	@Override
	public short getShort(int propertyIndex) {
		return (Short) get(propertyIndex);
	}

	@Override
	public byte[] getBytes(int propertyIndex) {
		return (byte[]) get(propertyIndex);
	}

	@Override
	public BigDecimal getBigDecimal(int propertyIndex) {
		return (BigDecimal) get(propertyIndex);
	}

	@Override
	public BigInteger getBigInteger(int propertyIndex) {
		return (BigInteger) get(propertyIndex);
	}

	@Override
	public DataObject getDataObject(int propertyIndex) {
		return (DataObject) get(propertyIndex);
	}

	@Override
	public Date getDate(int propertyIndex) {
		return (Date) get(propertyIndex);
	}

	@Override
	public String getString(int propertyIndex) {
		return get(propertyIndex).toString();
	}

	@Override
	public List<?> getList(int propertyIndex) {
		return (new DataObjectList(this, (Property) getType().getProperties().get(propertyIndex))).list(null);				
	}

	@Override
	public Sequence getSequence(int propertyIndex) {
		return (Sequence) get(propertyIndex);
	}

	@Override
	public void setBoolean(int propertyIndex, boolean value) {
		set(propertyIndex, value);
	}

	@Override
	public void setByte(int propertyIndex, byte value) {
		set(propertyIndex, value);
	}

	@Override
	public void setChar(int propertyIndex, char value) {
		set(propertyIndex, value);
	}

	@Override
	public void setDouble(int propertyIndex, double value) {
		set(propertyIndex, value);
	}

	@Override
	public void setFloat(int propertyIndex, float value) {
		set(propertyIndex, value);
	}

	@Override
	public void setInt(int propertyIndex, int value) {
		set(propertyIndex, value);
	}

	@Override
	public void setLong(int propertyIndex, long value) {
		set(propertyIndex, value);
	}

	@Override
	public void setShort(int propertyIndex, short value) {
		set(propertyIndex, value);
	}

	@Override
	public void setBytes(int propertyIndex, byte[] value) {
		set(propertyIndex, value);
	}

	@Override
	public void setBigDecimal(int propertyIndex, BigDecimal value) {
		set(propertyIndex, value);
	}

	@Override
	public void setBigInteger(int propertyIndex, BigInteger value) {
		set(propertyIndex, value);
	}

	@Override
	public void setDataObject(int propertyIndex, DataObject value) {
		set(propertyIndex, value);
	}

	@Override
	public void setDate(int propertyIndex, Date value) {
		set(propertyIndex, value);
	}

	@Override
	public void setString(int propertyIndex, String value) {
		set(propertyIndex, value);
	}

	@Override
	public void setList(int propertyIndex, List<?> value) {
		set(propertyIndex, value);
	}

	@Override
	public Object get(Property property) {
		if(this.instance == null)
			return null;

		Object value = ((ExtendedProperty)property).getValue(this.instance);
		if(value == null)
			return null;

		Object dataObject = objectCreator.getExistingDataObject(value);
		if(dataObject == null) {
			if(property.getType().isDataType())
				dataObject = value;
			else {
				ExtendedDataObject container = this;
				Property containmentProperty = property;
				if(!property.isContainment()) {
					container = null;
					containmentProperty = null;
				}
				dataObject = objectCreator.createDataObject(value, (ExtendedType) property.getType(), container, containmentProperty);
			}
		}

		return dataObject;
	}

	@Override
	public void set(Property property, Object value) {
		((ExtendedProperty)property).setValue(this.instance, value);
	}

	@Override
	public boolean isSet(Property property) {
		// TODO Auto-generated method stub
		return false;
	}

	@Override
	public void unset(Property property) {
		changeSummary.markModified(this, property, ((ExtendedProperty)property).getValue(this));
		set(property, null);
	}

	@Override
	public boolean getBoolean(Property property) {
		return (Boolean) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public byte getByte(Property property) {
		return (Byte) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public char getChar(Property property) {
		return (Character) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public double getDouble(Property property) {
		return (Double) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public float getFloat(Property property) {
		return (Float) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public int getInt(Property property) {
		return (Integer) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public long getLong(Property property) {
		return (Long) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public short getShort(Property property) {
		return (Short) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public byte[] getBytes(Property property) {
		return (byte[]) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public BigDecimal getBigDecimal(Property property) {
		return (BigDecimal) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public BigInteger getBigInteger(Property property) {
		return (BigInteger) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public DataObject getDataObject(Property property) {
		return objectCreator.getExistingDataObject(((ExtendedProperty)property).getValue(this));
	}

	@Override
	public Date getDate(Property property) {
		return (Date) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public String getString(Property property) {
		return ((ExtendedProperty)property).getValue(this).toString();
	}

	@Override
	public List<?> getList(Property property) {
		return (new DataObjectList(this, property)).list(null);					
	}

	@Override
	public Sequence getSequence(Property property) {
		return (Sequence) ((ExtendedProperty)property).getValue(this);
	}

	@Override
	public void setBoolean(Property property, boolean value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setByte(Property property, byte value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setChar(Property property, char value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setDouble(Property property, double value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setFloat(Property property, float value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setInt(Property property, int value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setLong(Property property, long value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setShort(Property property, short value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setBytes(Property property, byte[] value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setBigDecimal(Property property, BigDecimal value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setBigInteger(Property property, BigInteger value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setDataObject(Property property, DataObject value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setDate(Property property, Date value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setString(Property property, String value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public void setList(Property property, List<?> value) {
		((ExtendedProperty)property).setValue(this, value);
	}

	@Override
	public DataObject createDataObject(String propertyName) {
		Property property = getType().getProperty(propertyName);
		return createDataObject(property, property.getType());
	}

	@Override
	public DataObject createDataObject(int propertyIndex) {
		Property property = (Property) getType().getProperties().get(propertyIndex);
		return createDataObject(property, property.getType());
	}

	@Override
	public DataObject createDataObject(Property property) {
		return createDataObject(property, property.getType());
	}

	@Override
	public DataObject createDataObject(String propertyName,
			String namespaceURI, String typeName) {
		try {
			Class<?> javaClass = Class.forName(namespaceURI + typeName);
			DataAccessService das = ((ExtendedDataObject)getRootObject()).getObjectCreator().getDAS();
			Type type = das.getType(javaClass);

			Property property = getType().getProperty(propertyName);		
			return createDataObject(property, type);
		} catch (ClassNotFoundException e) {
			throw ClassUtil.wrapRun(e);
		}
	}

	@Override
	public DataObject createDataObject(int propertyIndex, String namespaceURI,
			String typeName) {

		try {
			Class<?> javaClass = Class.forName(namespaceURI + typeName);
			DataAccessService das = ((ExtendedDataObject)getRootObject()).getObjectCreator().getDAS();
			Type type = das.getType(javaClass);

			Property property = (Property) getType().getProperties().get(propertyIndex);			
			return createDataObject(property, type);
		} catch (ClassNotFoundException e) {
			throw ClassUtil.wrapRun(e);
		}
	}

	@Override
	public DataObject createDataObject(Property property, Type type) {
		if(instance == null)
			return null;

		if(property.getType().isDataType())
			throw new IllegalStateException("A DataObject needs to have properties");

		Object childInstance = ((ExtendedProperty)property).getValue(instance);
		ExtendedDataObject dataObject = ((ExtendedDataObject)getRootObject()).getObjectCreator().createDataObject(childInstance, (ExtendedType) type, this, property);

		return dataObject;
	}

	@Override
	public void delete() {
		changeSummary.markDeleted(this, getContainer(), getContainmentProperty());
	}

	@Override
	public DataObject getContainer() {
		return container;
	}

	@Override
	public void setContainer(DataObject value) {
		this.container = value;
	}

	@Override
	public void setContainmentProperty(Property value) {
		this.containmentProperty = value;
	}

	@Override
	public Property getContainmentProperty() {
		return containmentProperty;
	}

	@Override
	public DataGraph getDataGraph() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public Type getType() {
		return type;
	}

	@Override
	public Sequence getSequence() {
		return null;
	}

	@Override
	public List<?> getInstanceProperties() {
		return new ArrayList<Property>();
	}

	@Override
	public Property getInstanceProperty(String propertyName) {
		return getType().getProperty(propertyName);
	}

	@Override
	public Property getProperty(String propertyName) {
		return getInstanceProperty(propertyName);
	}

	@Override
	public DataObject getRootObject() {
		return (DataObject) get(PATH_DELIMITER);
	}

	@Override
	public ChangeSummary getChangeSummary() {
		return changeSummary;
	}

	@Override
	public void detach() {
		// TODO Auto-generated method stub

	}

	@Override
	public ObjectCreator getObjectCreator() {
		return objectCreator;
	}

	@Override
	public ExtendedDataObject load(Settings settings) {
		CallInfo callInfo = new CallInfo(this, null, null, null);
		settings.setAction(AggregateAction.UPDATE);
		callInfo.setSettings(settings);		

		ObjectCreator oc = new ObjectCreator(getObjectCreator().getDAS(), getObjectCreator().getPersistenceOrchestrator(), MapperDirection.REFERENCETOREFERENCE);
		return oc.createTarget(callInfo);
	}

	@Override
	public DataObject read(Settings settings) {
		// invoke read only on the root object
		if(!isRoot())
			throw new RuntimeException("Read needs to be invoked on the root data object");

		CallInfo callInfo = new CallInfo(this, null, null, null);
		settings.setAction(AggregateAction.READ);
		callInfo.setSettings(settings);		

		// Create an object creator for the target root
		ObjectCreator oc = new ObjectCreator(getObjectCreator().getDAS(), getObjectCreator().getPersistenceOrchestrator(), MapperDirection.DERIVEDTODERIVED);
		callInfo.setOutputObjectCreator(oc);
		ReadOperation operation = new ReadOperation();
		callInfo.setOperation(operation);
		Object target = operation.createTarget(callInfo, null);
		((ExtendedDataObject)target).setReadOnly(true);
		callInfo.setOutput(target);

		try {
			operation.execute(callInfo);
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}

		return (ExtendedDataObject) callInfo.getOutput();
	}

	@Override
	public List<?> query(Settings settings) {
		if(settings.getView() == null)
			throw new RuntimeException("A view needs to be specified to use the query method");

		CallInfo callInfo = new CallInfo(this, null, null, null);
		settings.setAction(AggregateAction.READ);
		callInfo.setSettings(settings);		

		ObjectCreator oc = new ObjectCreator(getObjectCreator().getDAS(), getObjectCreator().getPersistenceOrchestrator(), MapperDirection.DERIVEDTODERIVED);
		callInfo.setOutputObjectCreator(oc);		
		callInfo.setOperation(new QueryOperation());
		callInfo.setOutput(callInfo.getOperation().createTarget(callInfo, getObjectCreator().getDAS().getTypeMapper().toDerived(settings.getNarrowedClass())));
		((ExtendedDataObject)callInfo.getOutput()).setReadOnly(true);

		callInfo.getOperation().execute(callInfo);
		List<Object> result = (List<Object>) callInfo.getOperation().getResult();

		return result;
	}

	@Override
	public ExtendedDataObject update(Settings settings) {
		// clone the task object using a DataObject
		CallInfo callInfo = new CallInfo(this, null, null, null);
		callInfo.setSettings(settings);	
		callInfo.getSettings().setAction(settings.getAction()); // Since the default global action can be either UPDATE or MERGE
		this.createAggregate();

		// Create an object creator for the target root
		ObjectCreator oc = new ObjectCreator(getObjectCreator().getDAS(), getObjectCreator().getPersistenceOrchestrator(), MapperDirection.REFERENCETODERIVED);
		callInfo.setOutputObjectCreator(oc);
		ModifyOperation operation = new ModifyOperation();
		callInfo.setOperation(operation);
		Object target = operation.createTarget(callInfo, settings.getDesiredSubclass());
		callInfo.setOutput(target);

		try {
			operation.execute(callInfo);
			((ExtendedDataObject)target).getDependencyGraph().persistGraph(callInfo);
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}

		return (ExtendedDataObject) operation.getResult();
	}

	@Override
	/**
	 * Always creates a root data object
	 */
	public DataObject create(Settings settings) {

		// clone the task object using a DataObject
		CallInfo callInfo = new CallInfo(this, null, null, null);
		callInfo.setSettings(settings);	
		callInfo.getSettings().setAction(AggregateAction.CREATE);
		this.getObjectCreator().setShare(true);
		this.createAggregate();

		// Create an object creator for the target root
		ObjectCreator oc = new ObjectCreator(getObjectCreator().getDAS(), getObjectCreator().getPersistenceOrchestrator(), MapperDirection.REFERENCETODERIVED);
		callInfo.setOutputObjectCreator(oc);
		ModifyOperation operation = new ModifyOperation();
		callInfo.setOperation(operation);
		Object target = operation.createTarget(callInfo, settings.getDesiredSubclass());
		callInfo.setOutput(target);

		try {
			operation.execute(callInfo);
			((ExtendedDataObject)target).getDependencyGraph().persistGraph(callInfo);
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}

		return (ExtendedDataObject) operation.getResult();		
	}

	@Override
	public DataObject clone(Settings settings) {

		// clone the task object using a DataObject
		CallInfo callInfo = new CallInfo(this, null, null, null);
		callInfo.setSettings(settings);	
		callInfo.getSettings().setAction(AggregateAction.CLONE);
		this.createAggregate();

		// Create an object creator for the target root
		ObjectCreator oc = new ObjectCreator(getObjectCreator().getDAS(), getObjectCreator().getPersistenceOrchestrator(), MapperDirection.REFERENCETOREFERENCE);
		callInfo.setOutputObjectCreator(oc);
		CloneOperation operation = new CloneOperation();
		callInfo.setOperation(operation);
		Object target = operation.createTarget(callInfo, settings.getDesiredSubclass());
		callInfo.setOutput(target);

		try {
			operation.execute(callInfo);
			((ExtendedDataObject)target).getDependencyGraph().persistGraph(callInfo);
		} catch (Exception e) {
			throw ClassUtil.wrapRun(e);
		}

		return (ExtendedDataObject) operation.getResult();
	}	

	@Override
	public List<ExtendedDataObject> getList() {
		return (new DataObjectList(this)).list(null);
	}

	@Override
	public List<ExtendedDataObject> getList(CallInfo callInfo) {
		return (new DataObjectList(this)).list(callInfo);
	}	

	@Override
	public int hashCode() {
		int h = super.hashCode();

		if(getType().isDataType()) {
			h = 31 * h + instance.hashCode(); 
		}else { 
			ExtendedProperty identifierProperty = (ExtendedProperty) ((ExtendedType)type).getIdentifierProperty();
			Object id = identifierProperty.getValue(this);	

			if(id != null) {
				h = 31 * h + id.hashCode();
			} else {
				// iterate through the properties
				for(Property property: type.getProperties()) {
					ExtendedProperty extended = (ExtendedProperty) property;
					Object value = extended.getValue(this);
					h = 31 * h + value.hashCode();
				}
			}
		}

		return h;
	}

	@Override
	public boolean equals(Object o) {

		if(this == o)
			return true;

		if(!ExtendedDataObjectImpl.class.isAssignableFrom(o.getClass()))
			return false;

		ExtendedDataObjectImpl other = (ExtendedDataObjectImpl) o;
		if(getType() != other.getType() )
			return false;

		if(getType().isDataType())
			return instance.equals(other.getInstance());

		// Check if id is the same - then they are equal
		ExtendedProperty identifierProperty = (ExtendedProperty) ((ExtendedType)type).getIdentifierProperty();
		Object thisId = identifierProperty.getValue(this);
		Object otherId = identifierProperty.getValue(other);

		if(thisId != null)
			return thisId.equals(otherId);

		// Check each of the properties in turn
		for(Property property: type.getProperties()) {
			ExtendedProperty extended = (ExtendedProperty) property;
			Object thisValue = extended.getValue(this);
			Object otherValue = extended.getValue(other);
			if(thisValue != null && !thisValue.equals(otherValue))
				return false;
		}		

		return false;
	}

	protected boolean isRequiresValue(ExtendedDataObject obj, String propertyName, ContentDetail detail) {
		if(propertyName == null) // This object is part of a collection
			return isInDetail( obj.getInstance().getClass(), detail);

		if(isRequired(obj.getInstanceProperty(propertyName), detail))
			return true;

		return false;
	}	

	protected boolean isRequired(Property property, ContentDetail detail) {
		if(!property.isNullable())
			return true;

		// If this is not a required field then check the detail level to see if this field is required
		return isInDetail(property.getType().getClass(), detail);
	}	

	protected boolean isInDetail(Class<?> clazz, ContentDetail detail) {

		boolean result = false;

		// If this is not a required field then check the detail level to see if this field is required
		switch(detail) {
		case REQUIRED:
			break;
		case BASIC:
			return ClassUtil.isBasicType(clazz);
		case TEMPLATE:
		case FULL:
			return true;
		}

		return result;
	}	

	/**
	 * This method also takes case of invoking any business logic specific method
	 * @param copy
	 * @param propertyName
	 * @param propertyCopy
	 */
	protected void setCustomValue(CallInfo ci, ExtendedDataObject copy, String propertyName, Object propertyCopy) {

		// NOTE: pm.getCopy() may not be the same as the copy value passed in - e.g., the backRef code calling this methoed
		//       in the context of setting the bi-directional link

		AbstractProperty property = (AbstractProperty) copy.getInstanceProperty(propertyName);
		if(property.getCustomSetter() != null) {
			PropertyEvent event = new PropertyElement(ci, propertyCopy);
			property.setCustomValue(copy, event);
		} else
			property.setValue(copy, propertyCopy);		
	}

	@Override
	public Object getIdentifierValue() {
		if(instance == null)
			return null;

		if(!ExtendedType.class.isAssignableFrom(getType().getClass()))
			return null;

		ExtendedProperty idProperty = (ExtendedProperty) ((ExtendedType)getType()).getIdentifierProperty();
		if(idProperty == null)
			return null;

		return idProperty.getValue(this);
	}

	@Override
	public void checkInvariants() {
		((ExtendedType)this.getType()).checkInvariants(this.getInstance());
	}

	@Override
	public boolean isDependent() {
		boolean result = false;

		for(Edge edge: incoming) {
			if(edge.isCascaded() || edge.getProperty().isContainment())
				return true;
		}

		return result;
	}

	@Override
	public boolean isAggregateModified() {
		for(Edge edge: outgoing) {
			if(edge.isCascaded() && edge.target.isAggregateModified())
				return true;
		}

		// If none of its dependants are modified, return if the current object is modified
		return isModified();
	}

	@Override
	public void setAggregateName(String name) {
		this.aggregateName = name;
	}

	@Override
	public String getAggregateName() {
		return this.aggregateName;
	}

	@Override
	public boolean isEntity() {
		if(ExtendedType.class.isAssignableFrom(getType().getClass()) && ((ExtendedType)getType()).isEntity())
			return true;

		return false;
	}

	@Override
	public boolean isReferenceType() {
		return objectCreator.getTypeMapper().isReference(getType().getInstanceClass());
	}

	@Override
	public Type getReferenceType() {
		return getObjectCreator().getDAS().getType(getObjectCreator().getTypeMapper().toReference(getType().getInstanceClass()));
	}

	@Override
	public Type getDerivedType() {
		return getObjectCreator().getDAS().getType(getObjectCreator().getTypeMapper().toDerived(getType().getInstanceClass()));
	}

	@Override
	public void linkBackPointer() {
		for(Property p: getType().getProperties()) {	
			ExtendedProperty property = (ExtendedProperty) p;
			if(property.isContainment()) {
				if(property.getOpposite() != null)
					property.linkBackPointer(this);
				ExtendedDataObject value = (ExtendedDataObject)getDataObject(property);
				if(value != null)
					value.linkBackPointer();
			}
		}
	}

	@Override
	public void unlinkBackPointer() {
		for(Property p: getType().getProperties()) {	
			ExtendedProperty property = (ExtendedProperty) p;
			if(property.isContainment()) {
				if(property.getOpposite() != null)
					property.unlinkBackPointer(this);
				ExtendedDataObject value = (ExtendedDataObject)getDataObject(property);
				if(value != null)
					value.unlinkBackPointer();				
			}
		}
	}	
}
