/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton or third-party contributors as
 * indicated by the @author tags.
 *
 * This copyrighted material is made available to anyone wishing to use, modify,
 * copy, or redistribute it subject to the terms and conditions of the GNU
 * Lesser General Public License, as published by the Free Software Foundation.
 *
 * Ladder is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this distribution. If not, see <http://www.gnu.org/licenses/>.
 */

package ws.ladder.util.btree;

import java.util.HashMap;
import java.util.Map;

public abstract class AbstractStore implements Store {
	public static final int MIN_BLOCK_SIZE = 512;	

	protected int             blockSize;
	protected int             entrySize;
	protected IndexTypeEnum   type;
	protected Integer         minimumDegree;
	protected long            rootAddress;	
	protected Map<Long, Node> cachedNodes = new HashMap<Long, Node>();	

	public AbstractStore(int blockSize, int entrySize, IndexTypeEnum type) {
		this.blockSize = blockSize;
		this.entrySize = entrySize;
		this.type = type;
	}

	@Override
	public int getBlockSize() {
		return this.blockSize;
	}

	@Override
	public int getMinimumDegree() {
		/* 
		 * t is the minimum degree
		 * 
		 * - t >= 2
		 * - Every node other than the root must have at least t-1 keys. Every
		 *   internal node other than the root thus has at least t children.
		 *   If the tree is nonempty, the root must have at least one key.
		 * - Every node may contain at most 2t-1 keys. Therefore, an internal node
		 *   may have at most 2t children. We say that a node is full if it 
		 *   contains exactly 2t-1 keys.
		 */

		if(minimumDegree == null) {
			int dataLen = this.blockSize - Node.NODE_HEADER_LEN;
			int t = ((dataLen-Node.CHILD_PTR_LEN)/(Node.CHILD_PTR_LEN+entrySize) + 1)/2;
			
			if(t < 2)
				throw new RuntimeException("The minimum degree has to be >= 2");
			
			minimumDegree = t;
		}

		return minimumDegree;
	}	
	
	@Override
	public Node get(long address) {
		return cachedNodes.get(address);
	}	
	
	@Override
	public long getRootAddress() {
		return rootAddress;
	}	
}